#include <linux/module.h>
#include <rtai.h>
#include <rtai_sched.h>
#include <rtai_sem.h>
#include <rtai_fifos.h>

#define BASE_ADDRESS	0x378
#define DATA_REGISTER	(BASE_ADDRESS)
#define	STAT_REGISTER	(BASE_ADDRESS + 1)
#define CTRL_REGISTER	(BASE_ADDRESS + 2)

//#define DEBUG_LOCATIONS
//#define DEBUG_ACTIONS
//#define DEBUG_TRANSITIONS
//#define DEBUG_MAIN_LOOP
//#define DEBUG_EQUALS

#define NS_TIME_UNIT	5000000						// 5 millisec
#define NS_DELTA_P		(10000)						// Limite hardware (PAS PLUS BAS QUE 4000 !!!)
#define NS_DELTA_L		((NS_DELTA-4*NS_DELTA_P)/4)	// OK : DELTA > 3 DELTA_L + 4 DELTA_P
#define NS_DELTA		(NS_TIME_UNIT/5)			// OK : DELTA < U/4

#define TIME_UNIT	nano2count(NS_TIME_UNIT)
#define DELTA_P		nano2count(NS_DELTA_P)
#define DELTA_L		nano2count(NS_DELTA_L)
#define DELTA		nano2count(NS_DELTA)

static RT_TASK task;    

// AASAP-related ///////////////////////////////////////////////////////////////

static int AASAP_equals(RTIME a, RTIME b)
{
	RTIME lower = b - DELTA;
	RTIME higher = b + DELTA;
	if(a > lower && a < higher)
	{
		#ifdef DEBUG_EQUALS
			rt_printk("%d EQ %d with DELTA = %d\n", (int)a, (int)b, (int)DELTA);
		#endif
		return 1;
	}
	else
	{
		#ifdef DEBUG_EQUALS
			rt_printk("%d NEQ %d with DELTA = %d\n", (int)a, (int)b, (int)DELTA);
		#endif
		return 0;
	}
}

static int AASAP_biggerthan(RTIME a, RTIME other)
{
	other -= DELTA;
	if(a > other)
	{
		#ifdef DEBUG_EQUALS
			rt_printk("%d IS BIGGER THAN %d with DELTA = %d\n", (int)a, (int)other, (int)DELTA);
		#endif
		return 1;
	}
	else
	{
		#ifdef DEBUG_EQUALS
			rt_printk("%d IS NOT BIGGER THAN %d with DELTA = %d\n", (int)a, (int)other, (int)DELTA);
		#endif
		return 0;
	}
}



// Globals /////////////////////////////////////////////////////////////////////
static RTIME clock_x;
static unsigned char var_p, var_doublezero;
static void (*current_location)(void);
static int data_set[] = {
	1, 1, 1, 1, 1, 1, 1, 1,
	1, 0, 1, 0, 1, 0, 1, 1,
	1, 0, 1, 0, 1, 1, 1, 1,
	1, 0, 1, 0, 1, 0, 1, 1,
	1, 0, 1, 1, 1, 0, 1, 1,
	1, 0, 1, 0, 1, 0, 1, 1,
	1, 1, 1, 0, 1, 0, 1, 1,	
	1, 1, 1, 1, 1, 1, 1, 1,
	0, 0
};
static int data_set_length = 8*8+2;
static int data_set_index;

// Actions /////////////////////////////////////////////////////////////////////
static void action_up(void)
{
	#ifdef DEBUG_ACTIONS
		rt_printk("Action UP!\n");
	#endif
	outb(255, DATA_REGISTER);
}

static void action_down(void)
{
	#ifdef DEBUG_ACTIONS
		rt_printk("Action DOWN!\n");
	#endif
	outb(0, DATA_REGISTER);
}

static void action_inc_i(void)
{
	#ifdef DEBUG_ACTIONS
		rt_printk("Action INC!\n");
	#endif
	data_set_index++;
}

// Transitions /////////////////////////////////////////////////////////////////

// Declaration of location functions so the compiler doesn't complain...
static void location_idle(void);
static void location_onesent(void);
static void location_waitone(void);
static void location_zerosent(void);
static void location_waitzero(void);

////////////////////////////////////////////////////////////////////////////////

static void transition_init(void)
{
	#ifdef DEBUG_TRANSITIONS
		rt_printk(" -> init\n");
	#endif
	data_set_index = 0;
	clock_x = 0;
	var_p = 0;
	action_down();
	current_location = location_idle;
}

static void transition_idle_onesent(void)
{
	#ifdef DEBUG_TRANSITIONS
		rt_printk("Idle -> OneSent\n");
	#endif
	clock_x = 0;
	var_p = 0;
	var_doublezero = 0;
	action_up();
	action_inc_i();
	current_location = location_onesent;
}

static void transition_onesent_idle(void)
{
	#ifdef DEBUG_TRANSITIONS
		rt_printk("OneSent -> Idle\n");
	#endif
	clock_x = 0;
	var_p = 0;
	current_location = location_idle;
}

static void transition_onesent_waitone(void)
{
	#ifdef DEBUG_TRANSITIONS
		rt_printk("OneSent -> WaitOne\n");
	#endif
	clock_x = 0;
	action_down();
	current_location = location_waitone;
}

static void transition_onesent_zerosent(void)
{
	#ifdef DEBUG_TRANSITIONS
		rt_printk("OneSent -> ZeroSent\n");
	#endif
	clock_x = 0;
	var_p = var_p ? 0 : 1;

	action_down();
	action_inc_i();
	current_location = location_zerosent;
}

static void transition_waitone_onesent(void)
{
	#ifdef DEBUG_TRANSITIONS
		rt_printk("WaitOne -> OneSent\n");
	#endif
	clock_x = 0;
	var_p = var_p ? 0 : 1;
	action_up();
	action_inc_i();
	current_location = location_onesent;
}

static void transition_zerosent_waitzero(void)
{
	#ifdef DEBUG_TRANSITIONS
		rt_printk("ZeroSent -> WaitZero\n");
	#endif
	clock_x = 0;
	action_up();
	current_location = location_waitzero;
}

static void transition_waitzero_zerosent(void)
{
	#ifdef DEBUG_TRANSITIONS
		rt_printk("WaitZero -> ZeroSent\n");
	#endif
	clock_x = 0;
	var_p = var_p ? 0 : 1;
	var_doublezero = 1;
	action_down();
	action_inc_i();
	current_location = location_zerosent;
}

static void transition_zerosent_idle(void)
{
	#ifdef DEBUG_TRANSITIONS
		rt_printk("ZeroSent -> Idle\n");
	#endif
	clock_x = 0;
	var_p = 0;
	var_doublezero = 0;
	current_location = location_idle;
}

static void transition_zerosent_onesent(void)
{
	#ifdef DEBUG_TRANSITIONS
		rt_printk("ZeroSent -> OneSent\n");
	#endif
	clock_x = 0;
	var_p = var_p ? 0 : 1;
	var_doublezero = 0;
	action_up();
	action_inc_i();
	current_location = location_onesent;
}

// Locations ///////////////////////////////////////////////////////////////////
static void location_idle(void)
{
	#ifdef DEBUG_LOCATIONS
		rt_printk("Location Idle\n");
	#endif
	if (data_set_index == data_set_length)
		transition_init();
	if (AASAP_biggerthan(clock_x, (12*TIME_UNIT)))
		transition_idle_onesent();
}

static void location_onesent(void)
{
	#ifdef DEBUG_LOCATIONS
		rt_printk("Location OneSent\n");
	#endif
	if (data_set_index == data_set_length &&
		var_p == 1)
		transition_onesent_idle();

	if (data_set[data_set_index] == 1 &&
		AASAP_equals(clock_x, (2*TIME_UNIT)))
		transition_onesent_waitone();

	if (data_set[data_set_index] == 0 &&
		AASAP_equals(clock_x, (4*TIME_UNIT)))
		transition_onesent_zerosent();
}

static void location_waitone(void)
{
	#ifdef DEBUG_LOCATIONS
		rt_printk("Location Wait One\n");
	#endif
	if (AASAP_equals(clock_x, (2*TIME_UNIT)))
		transition_waitone_onesent();
}

static void location_zerosent(void)
{
	#ifdef DEBUG_LOCATIONS
		rt_printk("Location ZeroSent\n");
	#endif
	if (data_set_index == data_set_length &&
		( var_p == 1 || var_doublezero == 1))
		transition_zerosent_idle();

	if (data_set[data_set_index] == 0 &&
		AASAP_equals(clock_x, (2*TIME_UNIT)))
		transition_zerosent_waitzero();

	if (data_set[data_set_index] == 1 &&
		AASAP_equals(clock_x, (4*TIME_UNIT)))
		transition_zerosent_onesent();
}

static void location_waitzero(void)
{
	#ifdef DEBUG_LOCATIONS
		rt_printk("Location WaitZero\n");
	#endif
	if (AASAP_equals(clock_x, (2*TIME_UNIT)))
		transition_waitzero_zerosent();
}

// Main Loop ///////////////////////////////////////////////////////////////////

static void task_main(int data)
{
	RTIME time1 = rt_get_time(), time2;
	
	transition_init();

	while (1) {
		#ifdef DEBUG_MAIN_LOOP
			rt_printk("[%d] Index : %d \n", (int)clock_x, data_set_index);
		#endif
		current_location();
		rt_task_wait_period();
		time2 = rt_get_time();
		clock_x = clock_x + (time2 - time1);
		time1 = time2;
	}
}

////////////////////////////////////////////////////////////////////////////////


int xinit_module(void)
{
	RTIME rt_period;
	int stack_size = 10000;	
	int uses_fpu = 1;	

	outb(0, CTRL_REGISTER); //set port to output mode
    rt_set_periodic_mode();
	rt_period = DELTA_L;
	start_rt_timer(DELTA_P);
	
	if(DELTA_L > DELTA_P)
	{
		rt_task_init (
			&task,			// Structure RTAI RT_TASK
			task_main,		// Fonction main de la tâche
			0,				// Paramètre d'entrée
			stack_size,		// Taille du stack de la tâche
			0,				// Priorité de la tâche	(0 = plus prioritaire,
							//						 255 = moins prioritaire)
			uses_fpu,		// Flag d'utilisation du FPU
			0				// Callback du FIFO
		);
		
		rt_task_make_periodic(
			&task, 						// Structure RTAI RT_TASK
			rt_get_time() + rt_period,	// Offset
			rt_period					// Période
		);
	}
	else
		rt_printk("Let's not be suicidal...\n");

	return 0;
}

void xcleanup_module(void)
{
	stop_rt_timer();
	rt_task_delete(&task);
	return;
}

module_init(xinit_module);
module_exit(xcleanup_module);
MODULE_LICENSE("GPL");
