/* RTAI includes ------------------------------------------------------------ */
#include <linux/module.h>
#include <rtai.h>
#include <rtai_sched.h>
#include <rtai_sem.h>
#include <rtai_fifos.h>

/* user defines ------------------------------------------------------------- */
#define TIME_UNIT   <define this>
#define WIDENER     <define this>
#define PERIOD      <define this>
#define BEAT        <define this>

/* system defines ----------------------------------------------------------- */
#define STACK_SIZE 10000
#define USE_FPU 1
#define TASK_PRIORITY 128	/* 0 = highest, 255 = lowest */

/* global decoration -------------------------------------------------------- */

const int length = 8;
int var_a;
int array_b[length];
int b_index;


/* spectre gobals ----------------------------------------------------------- */
RTIME spectre_clock_x;

int spectre_equals (RTIME a, RTIME b);
int spectre_largerthan (RTIME a, RTIME other);
int spectre_smallerthan (RTIME a, RTIME other);
RT_TASK spectre_task;
void spectre_task_main (int data);
void (*spectre_current_location) (void);
RTIME spectre_current_time;
int spectre_running;
void spectre_check_for_events (void);
int spectre_e_is_pending = 0;


/* variables readings headers ----------------------------------------------- */
int spectre_read_a (void);

/* clocks readings headers -------------------------------------------------- */
RTIME spectre_read_x (void);


/* variables writings headers ----------------------------------------------- */
void spectre_write_a (int value);

/* clocks writings headers -------------------------------------------------- */
void spectre_write_x (RTIME value);


/* orders headers ----------------------------------------------------------- */
void spectre_d1_o (void);
void spectre_d2_o (void);

/* internals headers -------------------------------------------------------- */
void spectre_d1_i (void);
void spectre_d2_i (void);

/* events headers ----------------------------------------------------------- */
int spectre_poll_e (void);
void spectre_d1_e (void);
void spectre_d2_e (void);

/* location headers --------------------------------------------------------- */
void spectre_location_A (void);
void spectre_location_B (void);

/* transitions headers ------------------------------------------------------ */
void spectre_transition_initially (void);
void spectre_transition_A_to_B_1 (void);
void spectre_transition_A_to_B_2 (void);
void spectre_transition_B_to_B_1 (void);
void spectre_transition_B_to_A_2 (void);

/* updates headers ---------------------------------------------------------- */
void spectre_update_initially (void);
void spectre_update_A_to_B_1 (void);
void spectre_update_A_to_B_2 (void);
void spectre_update_B_to_B_1 (void);
void spectre_update_B_to_A_2 (void);

/* guards headers ----------------------------------------------------------- */
int spectre_guard_A_to_B_1 (void);
int spectre_guard_A_to_B_2 (void);
int spectre_guard_B_to_B_1 (void);
int spectre_guard_B_to_A_2 (void);

/* checks headers ----------------------------------------------------------- */
int spectre_check_A_to_B_1 (void);
int spectre_check_A_to_B_2 (void);
int spectre_check_B_to_B_1 (void);
int spectre_check_B_to_A_2 (void);

/* decorations headers ------------------------------------------------------ */
void spectre_d1_A_to_B (void);
void spectre_d1_B_to_A (void);
void spectre_d1_B_to_B (void);

void spectre_d2_A_to_B (void);
void spectre_d2_B_to_A (void);
void spectre_d2_B_to_B (void);

void spectre_d1_any_to_A (void);
void spectre_d1_any_to_B (void);

void spectre_d2_any_to_A (void);
void spectre_d2_any_to_B (void);

void spectre_d1_A_to_any (void);
void spectre_d1_B_to_any (void);

void spectre_d2_A_to_any (void);
void spectre_d2_B_to_any (void);

void spectre_d1_any_to_any (void);
void spectre_d2_any_to_any (void);


/* restrictions headers ----------------------------------------------------- */
int spectre_r_A_to_B (void);
int spectre_r_B_to_A (void);
int spectre_r_B_to_B (void);

int spectre_r_any_to_A (void);
int spectre_r_any_to_B (void);

int spectre_r_A_to_any (void);
int spectre_r_B_to_any (void);

int spectre_r_any_to_any (void);

/* variables readings implementation ---------------------------------------- */
int
spectre_read_a ()
{
  return var_a;
}

/* clocks readings implementation ------------------------------------------- */
RTIME
spectre_read_x (void)
{
  return spectre_current_time - spectre_clock_x;
}


/* variables writings implementation ---------------------------------------- */
void
spectre_write_a (int val)
{
  var_a = val;
}

/* clocks writings implementation ------------------------------------------- */
void
spectre_write_x (RTIME newtime)
{
  spectre_clock_x = spectre_current_time - newtime;
}


/* decorations implementation ----------------------------------------------- */
void
spectre_d1_A_to_B ()
{
  printf ("going from A to B !");
}

void
spectre_d1_B_to_A ()
{;
}

void
spectre_d1_B_to_B ()
{;
}

void
spectre_d2_A_to_B ()
{;
}

void
spectre_d2_B_to_A ()
{;
}

void
spectre_d2_B_to_B ()
{;
}

void
spectre_d1_any_to_A (void)
{;
}

void
spectre_d1_any_to_B (void)
{;
}

void
spectre_d2_any_to_A (void)
{;
}

void
spectre_d2_any_to_B (void)
{
  printf ("entering B !");
}

void
spectre_d1_A_to_any (void)
{
  printf ("leaving A !");
}

void
spectre_d1_B_to_any (void)
{;
}

void
spectre_d2_A_to_any (void)
{;
}

void
spectre_d2_B_to_any (void)
{;
}

void
spectre_d1_any_to_any (void)
{
  printf ("before update : i = %d", var_i);
}

void
spectre_d2_any_to_any (void)
{
  printf ("after update  : i = %d", var_i);
}


/* guards implementation ---------------------------------------------------- */
int
spectre_guard_One_to_Two_1 (void)
{
  RTIME spectre_t1, spectre_t2;
  int spectre_e1, spectre_e2;
  int spectre_res = 1, spectre_tmp;
  spectre_t1 = spectre_read_x ();
  spectre_t2 = (TIME_UNIT * 2);
  spectre_tmp = spectre_equals (spectre_t1, spectre_t2);
  spectre_res = spectre_tmp && spectre_res;
  return spectre_res;
}

int
spectre_guard_A_to_B_2 (void)
{
  RTIME spectre_t1, spectre_t2;
  int spectre_e1, spectre_e2;
  int spectre_res = 1, spectre_tmp;
  spectre_t1 = spectre_read_x ();
  spectre_t2 = (TIME_UNIT * 2);
  spectre_tmp = spectre_smallerthan (spectre_t1, spectre_t2);
  spectre_res = spectre_tmp && spectre_res;
  spectre_t1 = spectre_read_x ();
  spectre_t2 = (TIME_UNIT * 2);
  spectre_tmp = spectre_largerthan (spectre_t1, spectre_t2);
  spectre_res = spectre_tmp && spectre_res;
  return spectre_res;
}

int
spectre_guard_B_to_B_1 (void)
{
  RTIME spectre_t1, spectre_t2;
  int spectre_e1, spectre_e2;
  int spectre_res = 1, spectre_tmp;
  spectre_e1 = spectre_read_a ();
  spectre_e2 = 10;
  spectre_tmp = spectre_e1 <= spectre_e2;
  spectre_res = spectre_tmp && spectre_res;
  return spectre_res;
}

int
spectre_guard_B_to_A_2 (void)
{
  RTIME spectre_t1, spectre_t2;
  int spectre_e1, spectre_e2;
  int spectre_res = 1, spectre_tmp;
  spectre_e1 = spectre_read_a ();
  spectre_e2 = 10;
  spectre_tmp = spectre_e1 >= spectre_e2;
  spectre_res = spectre_tmp && spectre_res;
  return spectre_res;
}

/* restrictions implementation ---------------------------------------------- */
int
spectre_r_A_to_B (void)
{
  return 1;
}

int
spectre_r_B_to_A (void)
{
  return 1;
}

int
spectre_r_B_to_B (void)
{
  return 1;
}

int
spectre_r_any_to_A (void)
{
  return 1;
}

int
spectre_r_any_to_B (void)
{
  return 1;
}

int
spectre_r_A_to_any (void)
{
  return 1;
}

int
spectre_r_B_to_any (void)
{
  return 1;
}

int
spectre_r_any_to_any (void)
{
  return 1;
}

/* updates implementation --------------------------------------------------- */
void
spectre_update_initially (void)
{
  spectre_write_x ((TIME_UNIT * 0));
  spectre_write_a (0);
}

void
spectre_update_A_to_B_1 (void)
{
  spectre_write_x ((TIME_UNIT * 2));
  spectre_write_a (7);
}

void
spectre_update_A_to_B_2 (void)
{
  spectre_write_x ((TIME_UNIT * 2));
  spectre_write_a (spectre_read_a () + 4);
}

void
spectre_update_B_to_B_1 (void)
{
  spectre_write_a (spectre_read_a () + 1);
}

void
spectre_update_B_to_A_2 (void)
{
  spectre_write_a (0);
  spectre_write_x ((TIME_UNIT * 0));
}

/* orders implementation ---------------------------------------------------- */
void
spectre_d1_o (void)
{;
}

void
spectre_d2_o (void)
{
  printf ("emitting order o");
}

/* internals implementation ------------------------------------------------- */
void
spectre_d1_i (void)
{;
}

void
spectre_d2_i (void)
{
  printf ("treating internal i");
}

/* events implementation ---------------------------------------------------- */
void
spectre_d1_e (void)
{
  evt_has_been_treated ();
}

void
spectre_d2_e (void)
{;
}

int
spectre_poll_e (void)
{
  return check_presence_of_evt ();
}

/* checks implementation ---------------------------------------------------- */
int
spectre_check_A_to_B_1 (void)
{
  int spectre_r1, spectre_r2, spectre_r3, spectre_r4, spectre_g;
  spectre_r1 = spectre_r_A_to_B ();
  spectre_r2 = spectre_r_A_to_any ();
  spectre_r3 = spectre_r_any_to_B ();
  spectre_r4 = spectre_r_any_to_any ();
  spectre_g = spectre_guard_A_to_B_1 ();
  return (spectre_g && spectre_r1 && spectre_r2 && spectre_r3 && spectre_r4
	  && (spectre_e_is_pending == 1));
}

int
spectre_check_A_to_B_2 (void)
{
  int spectre_r1, spectre_r2, spectre_r3, spectre_r4, spectre_g;
  spectre_r1 = spectre_r_A_to_B ();
  spectre_r2 = spectre_r_A_to_any ();
  spectre_r3 = spectre_r_any_to_B ();
  spectre_r4 = spectre_r_any_to_any ();
  spectre_g = spectre_guard_A_to_B_2 ();
  return (spectre_g && spectre_r1 && spectre_r2 && spectre_r3 && spectre_r4
	  && (spectre_e_is_pending == 1));
}

int
spectre_check_B_to_B_1 (void)
{
  int spectre_r1, spectre_r2, spectre_r3, spectre_r4, spectre_g;
  spectre_r1 = spectre_r_B_to_B ();
  spectre_r2 = spectre_r_B_to_any ();
  spectre_r3 = spectre_r_any_to_B ();
  spectre_r4 = spectre_r_any_to_any ();
  spectre_g = spectre_guard_B_to_B_1 ();
  return (spectre_g && spectre_r1 && spectre_r2 && spectre_r3 && spectre_r4);
}

int
spectre_check_B_to_A_2 (void)
{
  int spectre_r1, spectre_r2, spectre_r3, spectre_r4, spectre_g;
  spectre_r1 = spectre_r_B_to_A ();
  spectre_r2 = spectre_r_B_to_any ();
  spectre_r3 = spectre_r_any_to_A ();
  spectre_r4 = spectre_r_any_to_any ();
  spectre_g = spectre_guard_B_to_A_2 ();
  return (spectre_g && spectre_r1 && spectre_r2 && spectre_r3 && spectre_r4);
}

/* transitions implementation ----------------------------------------------- */
void
spectre_transition_initially (void)
{
  spectre_update_initially ();
  spectre_current_location = spectre_location_A;
}

void spectre_transition_A_to_B_1 (void)
{
  spectre_d1_A_to_B ();
  spectre_d1_A_to_any ();
  spectre_d1_any_to_B ();
  spectre_d1_any_to_any ();
  spectre_d1_e ();
  spectre_update_A_to_B_1 ();
  spectre_current_location = spectre_location_B;
  spectre_e_is_pending = 0;
  spectre_d2_e ();
  spectre_d2_A_to_B ();
  spectre_d2_A_to_any ();
  spectre_d2_any_to_B ();
  spectre_d2_any_to_any ();
}

void
spectre_transition_A_to_B_2 (void)
{
  spectre_d1_A_to_B ();
  spectre_d1_A_to_any ();
  spectre_d1_any_to_B ();
  spectre_d1_any_to_any ();
  spectre_d1_e ();
  spectre_update_A_to_B_2 ();
  spectre_current_location = spectre_location_B;
  spectre_e_is_pending = 0;
  spectre_d2_e ();
  spectre_d2_A_to_B ();
  spectre_d2_A_to_any ();
  spectre_d2_any_to_B ();
  spectre_d2_any_to_any ();
}

void
spectre_transition_B_to_B_1 (void)
{
  spectre_d1_B_to_B ();
  spectre_d1_B_to_any ();
  spectre_d1_any_to_B ();
  spectre_d1_any_to_any ();
  spectre_d1_i ();
  spectre_update_B_to_B_1 ();
  spectre_d2_i ();
  spectre_d2_B_to_B ();
  spectre_d2_B_to_any ();
  spectre_d2_any_to_B ();
  spectre_d2_any_to_any ();
  spectre_current_location = spectre_location_B;
}

void
spectre_transition_B_to_A_2 (void)
{
  spectre_d1_B_to_A ();
  spectre_d1_B_to_any ();
  spectre_d1_any_to_A ();
  spectre_d1_any_to_any ();
  spectre_d1_o ();
  spectre_update_B_to_A_2 ();
  spectre_d2_o ();
  spectre_d2_B_to_A ();
  spectre_d2_B_to_any ();
  spectre_d2_any_to_A ();
  spectre_d2_any_to_any ();
  spectre_current_location = spectre_location_A;
}

/* location implementation -------------------------------------------------- */
void
spectre_location_A (void)
{
  if (spectre_check_A_to_B_1 ()) {
    spectre_transition_A_to_B_1 (); return;
  }

  if (spectre_check_A_to_B_2 ()) {
    spectre_transition_A_to_B_2 (); return;
  }
}

void
spectre_location_B (void)
{
  if (spectre_check_B_to_B_1 ())
    spectre_transition_B_to_B_1 ();
  if (spectre_check_B_to_A_2 ())
    spectre_transition_B_to_A_2 ();
}

void
spectre_check_for_events (void)
{
  if ((spectre_e_is_pending == 0) && spectre_poll_e ())
    spectre_e_is_pending = 1;
}

void
spectre_cleanup (void)
{ /* test cleanup */ ;
}

void
spectre_startup (void)
{ /* test startup */ ;
}

int
spectre_equals (RTIME a, RTIME b)
{
  RTIME lower = b - WIDENER;
  RTIME higher = b + WIDENER;
  return (a > lower && a < higher);
}

int
spectre_largerthan (RTIME a, RTIME other)
{
  other -= WIDENER;
  return (a > other);
}

int
spectre_smallerthan (RTIME a, RTIME other)
{
  other += WIDENER;
  return (a < other);
}

/* main function implementation --------------------------------------------- */
void
spectre_task_main (int data)
{
  spectre_running = 1;
  spectre_startup ();
  spectre_transition_initially ();
  while (spectre_running == 1)
    {
      spectre_current_time = rt_get_time ();
      spectre_check_for_events ();
      spectre_current_location ();
      rt_task_wait_period ();
    }
  spectre_cleanup ();
}

/* kernel module functions -------------------------------------------------- */
int
spectre_init_module (void)
{
  rt_set_periodic_mode ();
  start_rt_timer (BEAT);
  rt_task_init (&spectre_task, spectre_task_main, 0, STACK_SIZE,
		TASK_PRIORITY, USE_FPU, 0);
  rt_task_make_periodic (&spectre_task, rt_get_time () + PERIOD, PERIOD);
  return 0;
}

void
spectre_cleanup_module (void)
{
  stop_rt_timer ();
  rt_task_delete (&spectre_task);
}

module_init (spectre_init_module);
module_exit (spectre_cleanup_module);
MODULE_LICENSE ("GPL");
