/* RTAI includes ------------------------------------------------------------ */
#include <linux/module.h>
#include <rtai.h>
#include <rtai_sched.h>
#include <rtai_sem.h>
#include <rtai_fifos.h>

/* user defines ------------------------------------------------------------- */
#define TIME_UNIT   5000
#define WIDENER     201
#define PERIOD      200
#define BEAT        10

/* system defines ----------------------------------------------------------- */
#define STACK_SIZE 10000
#define USE_FPU 1
#define TASK_PRIORITY 128	/* 0 = highest, 255 = lowest */

/* global decoration -------------------------------------------------------- */

#define DEBUG

#define TRUE 1
#define FALSE 0

#define BASE_ADDRESS	0x378
#define DATA_REGISTER	(BASE_ADDRESS)
#define STAT_REGISTER	(BASE_ADDRESS + 1)
#define CTRL_REGISTER	(BASE_ADDRESS + 2)

int var_p, var_doublezero;
int data[] = {
  1, 1, 1, 1, 1, 1, 1, 1,
  1, 0, 1, 0, 1, 0, 1, 1,
  1, 0, 1, 0, 1, 1, 1, 1,
  1, 0, 1, 0, 1, 0, 1, 1,
  1, 0, 1, 1, 1, 0, 1, 1,
  1, 0, 1, 0, 1, 0, 1, 1,
  1, 1, 1, 0, 1, 0, 1, 1,
  1, 1, 1, 1, 1, 1, 1, 1,
  0, 0
};
const int data_length = 8 * 8 + 2;
int data_index;

int
sequence_finished (void)
{
  return data_index >= data_length;
}


/* spectre gobals ----------------------------------------------------------- */
RTIME spectre_clock_x;

int spectre_equals (RTIME a, RTIME b);
int spectre_largerthan (RTIME a, RTIME other);
int spectre_smallerthan (RTIME a, RTIME other);
RT_TASK spectre_task;
void spectre_task_main (int data);
void (*spectre_current_location) (void);
RTIME spectre_current_time;
int spectre_running;
void spectre_check_for_events (void);


/* variables readings headers ----------------------------------------------- */
int spectre_read_i (void);
int spectre_read_p (void);
int spectre_read_doublezero (void);

/* clocks readings headers -------------------------------------------------- */
RTIME spectre_read_x (void);


/* variables writings headers ----------------------------------------------- */
void spectre_write_i (int value);
void spectre_write_p (int value);
void spectre_write_doublezero (int value);

/* clocks writings headers -------------------------------------------------- */
void spectre_write_x (RTIME value);


/* orders headers ----------------------------------------------------------- */
void spectre_d1_up (void);
void spectre_d1_down (void);
void spectre_d2_up (void);
void spectre_d2_down (void);

/* internals headers -------------------------------------------------------- */

/* events headers ----------------------------------------------------------- */

/* location headers --------------------------------------------------------- */
void spectre_location_idle (void);
void spectre_location_oneSent (void);
void spectre_location_waitingOne (void);
void spectre_location_zeroSent (void);
void spectre_location_waitingZero (void);

/* transitions headers ------------------------------------------------------ */
void spectre_transition_initially (void);
void spectre_transition_idle_to_oneSent_1 (void);
void spectre_transition_oneSent_to_waitingOne_1 (void);
void spectre_transition_oneSent_to_zeroSent_2 (void);
void spectre_transition_oneSent_to_idle_3 (void);
void spectre_transition_waitingOne_to_oneSent_1 (void);
void spectre_transition_zeroSent_to_waitingZero_1 (void);
void spectre_transition_zeroSent_to_oneSent_2 (void);
void spectre_transition_zeroSent_to_idle_3 (void);
void spectre_transition_zeroSent_to_idle_4 (void);
void spectre_transition_waitingZero_to_zeroSent_1 (void);

/* updates headers ---------------------------------------------------------- */
void spectre_update_initially (void);
void spectre_update_idle_to_oneSent_1 (void);
void spectre_update_oneSent_to_waitingOne_1 (void);
void spectre_update_oneSent_to_zeroSent_2 (void);
void spectre_update_oneSent_to_idle_3 (void);
void spectre_update_waitingOne_to_oneSent_1 (void);
void spectre_update_zeroSent_to_waitingZero_1 (void);
void spectre_update_zeroSent_to_oneSent_2 (void);
void spectre_update_zeroSent_to_idle_3 (void);
void spectre_update_zeroSent_to_idle_4 (void);
void spectre_update_waitingZero_to_zeroSent_1 (void);

/* guards headers ----------------------------------------------------------- */
int spectre_guard_idle_to_oneSent_1 (void);
int spectre_guard_oneSent_to_waitingOne_1 (void);
int spectre_guard_oneSent_to_zeroSent_2 (void);
int spectre_guard_oneSent_to_idle_3 (void);
int spectre_guard_waitingOne_to_oneSent_1 (void);
int spectre_guard_zeroSent_to_waitingZero_1 (void);
int spectre_guard_zeroSent_to_oneSent_2 (void);
int spectre_guard_zeroSent_to_idle_3 (void);
int spectre_guard_zeroSent_to_idle_4 (void);
int spectre_guard_waitingZero_to_zeroSent_1 (void);

/* checks headers ----------------------------------------------------------- */
int spectre_check_idle_to_oneSent_1 (void);
int spectre_check_oneSent_to_waitingOne_1 (void);
int spectre_check_oneSent_to_zeroSent_2 (void);
int spectre_check_oneSent_to_idle_3 (void);
int spectre_check_waitingOne_to_oneSent_1 (void);
int spectre_check_zeroSent_to_waitingZero_1 (void);
int spectre_check_zeroSent_to_oneSent_2 (void);
int spectre_check_zeroSent_to_idle_3 (void);
int spectre_check_zeroSent_to_idle_4 (void);
int spectre_check_waitingZero_to_zeroSent_1 (void);

/* decorations headers ------------------------------------------------------ */
void spectre_d1_idle_to_oneSent (void);
void spectre_d1_oneSent_to_idle (void);
void spectre_d1_oneSent_to_waitingOne (void);
void spectre_d1_oneSent_to_zeroSent (void);
void spectre_d1_waitingOne_to_oneSent (void);
void spectre_d1_zeroSent_to_idle (void);
void spectre_d1_zeroSent_to_oneSent (void);
void spectre_d1_zeroSent_to_waitingZero (void);
void spectre_d1_waitingZero_to_zeroSent (void);

void spectre_d2_idle_to_oneSent (void);
void spectre_d2_oneSent_to_idle (void);
void spectre_d2_oneSent_to_waitingOne (void);
void spectre_d2_oneSent_to_zeroSent (void);
void spectre_d2_waitingOne_to_oneSent (void);
void spectre_d2_zeroSent_to_idle (void);
void spectre_d2_zeroSent_to_oneSent (void);
void spectre_d2_zeroSent_to_waitingZero (void);
void spectre_d2_waitingZero_to_zeroSent (void);

void spectre_d1_any_to_idle (void);
void spectre_d1_any_to_oneSent (void);
void spectre_d1_any_to_waitingOne (void);
void spectre_d1_any_to_zeroSent (void);
void spectre_d1_any_to_waitingZero (void);

void spectre_d2_any_to_idle (void);
void spectre_d2_any_to_oneSent (void);
void spectre_d2_any_to_waitingOne (void);
void spectre_d2_any_to_zeroSent (void);
void spectre_d2_any_to_waitingZero (void);

void spectre_d1_idle_to_any (void);
void spectre_d1_oneSent_to_any (void);
void spectre_d1_waitingOne_to_any (void);
void spectre_d1_zeroSent_to_any (void);
void spectre_d1_waitingZero_to_any (void);

void spectre_d2_idle_to_any (void);
void spectre_d2_oneSent_to_any (void);
void spectre_d2_waitingOne_to_any (void);
void spectre_d2_zeroSent_to_any (void);
void spectre_d2_waitingZero_to_any (void);

void spectre_d1_any_to_any (void);
void spectre_d2_any_to_any (void);


/* restrictions headers ----------------------------------------------------- */
int spectre_r_idle_to_oneSent (void);
int spectre_r_oneSent_to_idle (void);
int spectre_r_oneSent_to_waitingOne (void);
int spectre_r_oneSent_to_zeroSent (void);
int spectre_r_waitingOne_to_oneSent (void);
int spectre_r_zeroSent_to_idle (void);
int spectre_r_zeroSent_to_oneSent (void);
int spectre_r_zeroSent_to_waitingZero (void);
int spectre_r_waitingZero_to_zeroSent (void);

int spectre_r_any_to_idle (void);
int spectre_r_any_to_oneSent (void);
int spectre_r_any_to_waitingOne (void);
int spectre_r_any_to_zeroSent (void);
int spectre_r_any_to_waitingZero (void);

int spectre_r_idle_to_any (void);
int spectre_r_oneSent_to_any (void);
int spectre_r_waitingOne_to_any (void);
int spectre_r_zeroSent_to_any (void);
int spectre_r_waitingZero_to_any (void);

int spectre_r_any_to_any (void);

/* variables readings implementation ---------------------------------------- */
int
spectre_read_i ()
{
#ifdef DEBUG
  rt_printk ("reading variable i\n");
#endif
  return data[data_index];
}

int
spectre_read_p ()
{
#ifdef DEBUG
  rt_printk ("reading variable p\n");
#endif
  return var_p;
}

int
spectre_read_doublezero ()
{
#ifdef DEBUG
  rt_printk ("reading variable doublezero");
#endif
  return var_doublezero;
}

/* clocks readings implementation ------------------------------------------- */
RTIME
spectre_read_x (void)
{
  return spectre_current_time - spectre_clock_x;
}


/* variables writings implementation ---------------------------------------- */
void
spectre_write_i (int foo)
{;
}

void
spectre_write_p (int value)
{
#ifdef DEBUG
  rt_printk ("writing variable p\n");
#endif
  var_p = value;
}

void
spectre_write_doublezero (int value)
{
#ifdef DEBUG
  rt_printk ("Writing variable doublezero\n");
#endif
  var_doublezero = value;
}

/* clocks writings implementation ------------------------------------------- */
void
spectre_write_x (RTIME newtime)
{
  spectre_clock_x = spectre_current_time - newtime;
}


/* decorations implementation ----------------------------------------------- */
void
spectre_d1_idle_to_oneSent ()
{;
}

void
spectre_d1_oneSent_to_idle ()
{;
}

void
spectre_d1_oneSent_to_waitingOne ()
{;
}

void
spectre_d1_oneSent_to_zeroSent ()
{;
}

void
spectre_d1_waitingOne_to_oneSent ()
{;
}

void
spectre_d1_zeroSent_to_idle ()
{;
}

void
spectre_d1_zeroSent_to_oneSent ()
{;
}

void
spectre_d1_zeroSent_to_waitingZero ()
{;
}

void
spectre_d1_waitingZero_to_zeroSent ()
{;
}

void
spectre_d2_idle_to_oneSent ()
{;
}

void
spectre_d2_oneSent_to_idle ()
{;
}

void
spectre_d2_oneSent_to_waitingOne ()
{;
}

void
spectre_d2_oneSent_to_zeroSent ()
{;
}

void
spectre_d2_waitingOne_to_oneSent ()
{;
}

void
spectre_d2_zeroSent_to_idle ()
{;
}

void
spectre_d2_zeroSent_to_oneSent ()
{;
}

void
spectre_d2_zeroSent_to_waitingZero ()
{;
}

void
spectre_d2_waitingZero_to_zeroSent ()
{;
}

void
spectre_d1_any_to_idle (void)
{;
}

void
spectre_d1_any_to_oneSent (void)
{;
}

void
spectre_d1_any_to_waitingOne (void)
{;
}

void
spectre_d1_any_to_zeroSent (void)
{;
}

void
spectre_d1_any_to_waitingZero (void)
{;
}

void
spectre_d2_any_to_idle (void)
{
#ifdef DEBUG
  rt_printk ("entering idle\n");
#endif
  spectre_running = 0;
}

void
spectre_d2_any_to_oneSent (void)
{
#ifdef DEBUG
  rt_printk ("entering oneSent\n");
#endif
  data_index++;
}

void
spectre_d2_any_to_waitingOne (void)
{;
}

void
spectre_d2_any_to_zeroSent (void)
{
#ifdef DEBUG
  rt_printk ("entering zeroSent\n");
#endif
  data_index++;
}

void
spectre_d2_any_to_waitingZero (void)
{;
}

void
spectre_d1_idle_to_any (void)
{;
}

void
spectre_d1_oneSent_to_any (void)
{;
}

void
spectre_d1_waitingOne_to_any (void)
{;
}

void
spectre_d1_zeroSent_to_any (void)
{;
}

void
spectre_d1_waitingZero_to_any (void)
{;
}

void
spectre_d2_idle_to_any (void)
{
  data_index = 0;
}

void
spectre_d2_oneSent_to_any (void)
{;
}

void
spectre_d2_waitingOne_to_any (void)
{;
}

void
spectre_d2_zeroSent_to_any (void)
{;
}

void
spectre_d2_waitingZero_to_any (void)
{;
}

void
spectre_d1_any_to_any (void)
{;
}

void
spectre_d2_any_to_any (void)
{;
}


/* guards implementation ---------------------------------------------------- */
int
spectre_guard_idle_to_oneSent_1 (void)
{
  RTIME spectre_t1, spectre_t2;
  int spectre_e1, spectre_e2;
  int spectre_res = 1, spectre_tmp;
  spectre_t1 = spectre_read_x ();
  spectre_t2 = (TIME_UNIT * 12);
  spectre_tmp = spectre_largerthan (spectre_t1, spectre_t2);
  spectre_res = spectre_tmp && spectre_res;
  return spectre_res;
}

int
spectre_guard_oneSent_to_waitingOne_1 (void)
{
  RTIME spectre_t1, spectre_t2;
  int spectre_e1, spectre_e2;
  int spectre_res = 1, spectre_tmp;
  spectre_t1 = spectre_read_x ();
  spectre_t2 = (TIME_UNIT * 2);
  spectre_tmp = spectre_equals (spectre_t1, spectre_t2);
  spectre_res = spectre_tmp && spectre_res;
  spectre_e1 = spectre_read_i ();
  spectre_e2 = 1;
  spectre_tmp = spectre_e1 == spectre_e2;
  spectre_res = spectre_tmp && spectre_res;
  return spectre_res;
}

int
spectre_guard_oneSent_to_zeroSent_2 (void)
{
  RTIME spectre_t1, spectre_t2;
  int spectre_e1, spectre_e2;
  int spectre_res = 1, spectre_tmp;
  spectre_t1 = spectre_read_x ();
  spectre_t2 = (TIME_UNIT * 4);
  spectre_tmp = spectre_equals (spectre_t1, spectre_t2);
  spectre_res = spectre_tmp && spectre_res;
  spectre_e1 = spectre_read_i ();
  spectre_e2 = 0;
  spectre_tmp = spectre_e1 == spectre_e2;
  spectre_res = spectre_tmp && spectre_res;
  return spectre_res;
}

int
spectre_guard_oneSent_to_idle_3 (void)
{
  RTIME spectre_t1, spectre_t2;
  int spectre_e1, spectre_e2;
  int spectre_res = 1, spectre_tmp;
  spectre_e1 = spectre_read_p ();
  spectre_e2 = 1;
  spectre_tmp = spectre_e1 == spectre_e2;
  spectre_res = spectre_tmp && spectre_res;
  return spectre_res;
}

int
spectre_guard_waitingOne_to_oneSent_1 (void)
{
  RTIME spectre_t1, spectre_t2;
  int spectre_e1, spectre_e2;
  int spectre_res = 1, spectre_tmp;
  spectre_t1 = spectre_read_x ();
  spectre_t2 = (TIME_UNIT * 2);
  spectre_tmp = spectre_equals (spectre_t1, spectre_t2);
  spectre_res = spectre_tmp && spectre_res;
  return spectre_res;
}

int
spectre_guard_zeroSent_to_waitingZero_1 (void)
{
  RTIME spectre_t1, spectre_t2;
  int spectre_e1, spectre_e2;
  int spectre_res = 1, spectre_tmp;
  spectre_t1 = spectre_read_x ();
  spectre_t2 = (TIME_UNIT * 2);
  spectre_tmp = spectre_equals (spectre_t1, spectre_t2);
  spectre_res = spectre_tmp && spectre_res;
  spectre_e1 = spectre_read_i ();
  spectre_e2 = 0;
  spectre_tmp = spectre_e1 == spectre_e2;
  spectre_res = spectre_tmp && spectre_res;
  return spectre_res;
}

int
spectre_guard_zeroSent_to_oneSent_2 (void)
{
  RTIME spectre_t1, spectre_t2;
  int spectre_e1, spectre_e2;
  int spectre_res = 1, spectre_tmp;
  spectre_t1 = spectre_read_x ();
  spectre_t2 = (TIME_UNIT * 4);
  spectre_tmp = spectre_equals (spectre_t1, spectre_t2);
  spectre_res = spectre_tmp && spectre_res;
  spectre_e1 = spectre_read_i ();
  spectre_e2 = 1;
  spectre_tmp = spectre_e1 == spectre_e2;
  spectre_res = spectre_tmp && spectre_res;
  return spectre_res;
}

int
spectre_guard_zeroSent_to_idle_3 (void)
{
  RTIME spectre_t1, spectre_t2;
  int spectre_e1, spectre_e2;
  int spectre_res = 1, spectre_tmp;
  spectre_e1 = spectre_read_p ();
  spectre_e2 = 1;
  spectre_tmp = spectre_e1 == spectre_e2;
  spectre_res = spectre_tmp && spectre_res;
  return spectre_res;
}

int
spectre_guard_zeroSent_to_idle_4 (void)
{
  RTIME spectre_t1, spectre_t2;
  int spectre_e1, spectre_e2;
  int spectre_res = 1, spectre_tmp;
  spectre_e1 = spectre_read_doublezero ();
  spectre_e2 = 1;
  spectre_tmp = spectre_e1 == spectre_e2;
  spectre_res = spectre_tmp && spectre_res;
  return spectre_res;
}

int
spectre_guard_waitingZero_to_zeroSent_1 (void)
{
  RTIME spectre_t1, spectre_t2;
  int spectre_e1, spectre_e2;
  int spectre_res = 1, spectre_tmp;
  spectre_t1 = spectre_read_x ();
  spectre_t2 = (TIME_UNIT * 2);
  spectre_tmp = spectre_equals (spectre_t1, spectre_t2);
  spectre_res = spectre_tmp && spectre_res;
  return spectre_res;
}

/* restrictions implementation ---------------------------------------------- */
int
spectre_r_idle_to_oneSent (void)
{
  return 1;
}

int
spectre_r_oneSent_to_idle (void)
{
  return 1;
}

int
spectre_r_oneSent_to_waitingOne (void)
{
  return !sequence_finished ();
}

int
spectre_r_oneSent_to_zeroSent (void)
{
  return 1;
}

int
spectre_r_waitingOne_to_oneSent (void)
{
  return 1;
}

int
spectre_r_zeroSent_to_idle (void)
{
  return 1;
}

int
spectre_r_zeroSent_to_oneSent (void)
{
  return 1;
}

int
spectre_r_zeroSent_to_waitingZero (void)
{
  return !sequence_finished ();
}

int
spectre_r_waitingZero_to_zeroSent (void)
{
  return 1;
}

int
spectre_r_any_to_idle (void)
{
  return sequence_finished ();
}

int
spectre_r_any_to_oneSent (void)
{
  return 1;
}

int
spectre_r_any_to_waitingOne (void)
{
  return 1;
}

int
spectre_r_any_to_zeroSent (void)
{
  return 1;
}

int
spectre_r_any_to_waitingZero (void)
{
  return 1;
}

int
spectre_r_idle_to_any (void)
{
  return 1;
}

int
spectre_r_oneSent_to_any (void)
{
  return 1;
}

int
spectre_r_waitingOne_to_any (void)
{
  return 1;
}

int
spectre_r_zeroSent_to_any (void)
{
  return 1;
}

int
spectre_r_waitingZero_to_any (void)
{
  return 1;
}

int
spectre_r_any_to_any (void)
{
  return 1;
}

/* updates implementation --------------------------------------------------- */
void
spectre_update_initially (void)
{
  spectre_write_x ((TIME_UNIT * 0));
  spectre_write_p (0);
  spectre_write_doublezero (0);
}

void
spectre_update_idle_to_oneSent_1 (void)
{
  spectre_write_x ((TIME_UNIT * 0));
  spectre_write_p (0);
  spectre_write_doublezero (0);
}

void
spectre_update_oneSent_to_waitingOne_1 (void)
{
  spectre_write_x ((TIME_UNIT * 0));
}

void
spectre_update_oneSent_to_zeroSent_2 (void)
{
  spectre_write_x ((TIME_UNIT * 0));
  spectre_write_p (1 - spectre_read_p ());
}

void
spectre_update_oneSent_to_idle_3 (void)
{
  spectre_write_x ((TIME_UNIT * 0));
  spectre_write_p (0);
}

void
spectre_update_waitingOne_to_oneSent_1 (void)
{
  spectre_write_x ((TIME_UNIT * 0));
  spectre_write_p (1 - spectre_read_p ());
}

void
spectre_update_zeroSent_to_waitingZero_1 (void)
{
  spectre_write_x ((TIME_UNIT * 0));
}

void
spectre_update_zeroSent_to_oneSent_2 (void)
{
  spectre_write_x ((TIME_UNIT * 0));
  spectre_write_p (1 - spectre_read_p ());
  spectre_write_doublezero (0);
}

void
spectre_update_zeroSent_to_idle_3 (void)
{
  spectre_write_x ((TIME_UNIT * 0));
  spectre_write_p (0);
  spectre_write_doublezero (0);
}

void
spectre_update_zeroSent_to_idle_4 (void)
{
  spectre_write_x ((TIME_UNIT * 0));
  spectre_write_p (0);
  spectre_write_doublezero (0);
}

void
spectre_update_waitingZero_to_zeroSent_1 (void)
{
  spectre_write_x ((TIME_UNIT * 0));
  spectre_write_p (1 - spectre_read_p ());
  spectre_write_doublezero (1);
}

/* orders implementation ---------------------------------------------------- */
void
spectre_d1_up (void)
{
#ifdef DEBUG
  rt_printk ("Order UP!\n");
#endif
  outb (255, DATA_REGISTER);
}

void
spectre_d1_down (void)
{
#ifdef DEBUG
  rt_printk ("Order DOWN!\n");
#endif
  outb (0, DATA_REGISTER);
}

void
spectre_d2_up (void)
{;
}

void
spectre_d2_down (void)
{;
}

/* internals implementation ------------------------------------------------- */

/* events implementation ---------------------------------------------------- */

/* checks implementation ---------------------------------------------------- */
int
spectre_check_idle_to_oneSent_1 (void)
{
  int spectre_r1, spectre_r2, spectre_r3, spectre_r4, spectre_g;
  spectre_r1 = spectre_r_idle_to_oneSent ();
  spectre_r2 = spectre_r_idle_to_any ();
  spectre_r3 = spectre_r_any_to_oneSent ();
  spectre_r4 = spectre_r_any_to_any ();
  spectre_g = spectre_guard_idle_to_oneSent_1 ();
  return (spectre_g && spectre_r1 && spectre_r2 && spectre_r3 && spectre_r4);
}

int
spectre_check_oneSent_to_waitingOne_1 (void)
{
  int spectre_r1, spectre_r2, spectre_r3, spectre_r4, spectre_g;
  spectre_r1 = spectre_r_oneSent_to_waitingOne ();
  spectre_r2 = spectre_r_oneSent_to_any ();
  spectre_r3 = spectre_r_any_to_waitingOne ();
  spectre_r4 = spectre_r_any_to_any ();
  spectre_g = spectre_guard_oneSent_to_waitingOne_1 ();
  return (spectre_g && spectre_r1 && spectre_r2 && spectre_r3 && spectre_r4);
}

int
spectre_check_oneSent_to_zeroSent_2 (void)
{
  int spectre_r1, spectre_r2, spectre_r3, spectre_r4, spectre_g;
  spectre_r1 = spectre_r_oneSent_to_zeroSent ();
  spectre_r2 = spectre_r_oneSent_to_any ();
  spectre_r3 = spectre_r_any_to_zeroSent ();
  spectre_r4 = spectre_r_any_to_any ();
  spectre_g = spectre_guard_oneSent_to_zeroSent_2 ();
  return (spectre_g && spectre_r1 && spectre_r2 && spectre_r3 && spectre_r4);
}

int
spectre_check_oneSent_to_idle_3 (void)
{
  int spectre_r1, spectre_r2, spectre_r3, spectre_r4, spectre_g;
  spectre_r1 = spectre_r_oneSent_to_idle ();
  spectre_r2 = spectre_r_oneSent_to_any ();
  spectre_r3 = spectre_r_any_to_idle ();
  spectre_r4 = spectre_r_any_to_any ();
  spectre_g = spectre_guard_oneSent_to_idle_3 ();
  return (spectre_g && spectre_r1 && spectre_r2 && spectre_r3 && spectre_r4);
}

int
spectre_check_waitingOne_to_oneSent_1 (void)
{
  int spectre_r1, spectre_r2, spectre_r3, spectre_r4, spectre_g;
  spectre_r1 = spectre_r_waitingOne_to_oneSent ();
  spectre_r2 = spectre_r_waitingOne_to_any ();
  spectre_r3 = spectre_r_any_to_oneSent ();
  spectre_r4 = spectre_r_any_to_any ();
  spectre_g = spectre_guard_waitingOne_to_oneSent_1 ();
  return (spectre_g && spectre_r1 && spectre_r2 && spectre_r3 && spectre_r4);
}

int
spectre_check_zeroSent_to_waitingZero_1 (void)
{
  int spectre_r1, spectre_r2, spectre_r3, spectre_r4, spectre_g;
  spectre_r1 = spectre_r_zeroSent_to_waitingZero ();
  spectre_r2 = spectre_r_zeroSent_to_any ();
  spectre_r3 = spectre_r_any_to_waitingZero ();
  spectre_r4 = spectre_r_any_to_any ();
  spectre_g = spectre_guard_zeroSent_to_waitingZero_1 ();
  return (spectre_g && spectre_r1 && spectre_r2 && spectre_r3 && spectre_r4);
}

int
spectre_check_zeroSent_to_oneSent_2 (void)
{
  int spectre_r1, spectre_r2, spectre_r3, spectre_r4, spectre_g;
  spectre_r1 = spectre_r_zeroSent_to_oneSent ();
  spectre_r2 = spectre_r_zeroSent_to_any ();
  spectre_r3 = spectre_r_any_to_oneSent ();
  spectre_r4 = spectre_r_any_to_any ();
  spectre_g = spectre_guard_zeroSent_to_oneSent_2 ();
  return (spectre_g && spectre_r1 && spectre_r2 && spectre_r3 && spectre_r4);
}

int
spectre_check_zeroSent_to_idle_3 (void)
{
  int spectre_r1, spectre_r2, spectre_r3, spectre_r4, spectre_g;
  spectre_r1 = spectre_r_zeroSent_to_idle ();
  spectre_r2 = spectre_r_zeroSent_to_any ();
  spectre_r3 = spectre_r_any_to_idle ();
  spectre_r4 = spectre_r_any_to_any ();
  spectre_g = spectre_guard_zeroSent_to_idle_3 ();
  return (spectre_g && spectre_r1 && spectre_r2 && spectre_r3 && spectre_r4);
}

int
spectre_check_zeroSent_to_idle_4 (void)
{
  int spectre_r1, spectre_r2, spectre_r3, spectre_r4, spectre_g;
  spectre_r1 = spectre_r_zeroSent_to_idle ();
  spectre_r2 = spectre_r_zeroSent_to_any ();
  spectre_r3 = spectre_r_any_to_idle ();
  spectre_r4 = spectre_r_any_to_any ();
  spectre_g = spectre_guard_zeroSent_to_idle_4 ();
  return (spectre_g && spectre_r1 && spectre_r2 && spectre_r3 && spectre_r4);
}

int
spectre_check_waitingZero_to_zeroSent_1 (void)
{
  int spectre_r1, spectre_r2, spectre_r3, spectre_r4, spectre_g;
  spectre_r1 = spectre_r_waitingZero_to_zeroSent ();
  spectre_r2 = spectre_r_waitingZero_to_any ();
  spectre_r3 = spectre_r_any_to_zeroSent ();
  spectre_r4 = spectre_r_any_to_any ();
  spectre_g = spectre_guard_waitingZero_to_zeroSent_1 ();
  return (spectre_g && spectre_r1 && spectre_r2 && spectre_r3 && spectre_r4);
}

/* transitions implementation ----------------------------------------------- */
void
spectre_transition_initially (void)
{
  spectre_update_initially ();
  spectre_current_location = spectre_location_idle;
}

void
spectre_transition_idle_to_oneSent_1 (void)
{
  spectre_d1_idle_to_oneSent ();
  spectre_d1_idle_to_any ();
  spectre_d1_any_to_oneSent ();
  spectre_d1_any_to_any ();
  spectre_d1_up ();
  spectre_update_idle_to_oneSent_1 ();
  spectre_d2_up ();
  spectre_d2_idle_to_oneSent ();
  spectre_d2_idle_to_any ();
  spectre_d2_any_to_oneSent ();
  spectre_d2_any_to_any ();
  spectre_current_location = spectre_location_oneSent;
}

void
spectre_transition_oneSent_to_waitingOne_1 (void)
{
  spectre_d1_oneSent_to_waitingOne ();
  spectre_d1_oneSent_to_any ();
  spectre_d1_any_to_waitingOne ();
  spectre_d1_any_to_any ();
  spectre_d1_down ();
  spectre_update_oneSent_to_waitingOne_1 ();
  spectre_d2_down ();
  spectre_d2_oneSent_to_waitingOne ();
  spectre_d2_oneSent_to_any ();
  spectre_d2_any_to_waitingOne ();
  spectre_d2_any_to_any ();
  spectre_current_location = spectre_location_waitingOne;
}

void
spectre_transition_oneSent_to_zeroSent_2 (void)
{
  spectre_d1_oneSent_to_zeroSent ();
  spectre_d1_oneSent_to_any ();
  spectre_d1_any_to_zeroSent ();
  spectre_d1_any_to_any ();
  spectre_d1_down ();
  spectre_update_oneSent_to_zeroSent_2 ();
  spectre_d2_down ();
  spectre_d2_oneSent_to_zeroSent ();
  spectre_d2_oneSent_to_any ();
  spectre_d2_any_to_zeroSent ();
  spectre_d2_any_to_any ();
  spectre_current_location = spectre_location_zeroSent;
}

void
spectre_transition_oneSent_to_idle_3 (void)
{
  spectre_d1_oneSent_to_idle ();
  spectre_d1_oneSent_to_any ();
  spectre_d1_any_to_idle ();
  spectre_d1_any_to_any ();
  spectre_update_oneSent_to_idle_3 ();
  spectre_d2_oneSent_to_idle ();
  spectre_d2_oneSent_to_any ();
  spectre_d2_any_to_idle ();
  spectre_d2_any_to_any ();
  spectre_current_location = spectre_location_idle;
}

void
spectre_transition_waitingOne_to_oneSent_1 (void)
{
  spectre_d1_waitingOne_to_oneSent ();
  spectre_d1_waitingOne_to_any ();
  spectre_d1_any_to_oneSent ();
  spectre_d1_any_to_any ();
  spectre_d1_up ();
  spectre_update_waitingOne_to_oneSent_1 ();
  spectre_d2_up ();
  spectre_d2_waitingOne_to_oneSent ();
  spectre_d2_waitingOne_to_any ();
  spectre_d2_any_to_oneSent ();
  spectre_d2_any_to_any ();
  spectre_current_location = spectre_location_oneSent;
}

void
spectre_transition_zeroSent_to_waitingZero_1 (void)
{
  spectre_d1_zeroSent_to_waitingZero ();
  spectre_d1_zeroSent_to_any ();
  spectre_d1_any_to_waitingZero ();
  spectre_d1_any_to_any ();
  spectre_d1_up ();
  spectre_update_zeroSent_to_waitingZero_1 ();
  spectre_d2_up ();
  spectre_d2_zeroSent_to_waitingZero ();
  spectre_d2_zeroSent_to_any ();
  spectre_d2_any_to_waitingZero ();
  spectre_d2_any_to_any ();
  spectre_current_location = spectre_location_waitingZero;
}

void
spectre_transition_zeroSent_to_oneSent_2 (void)
{
  spectre_d1_zeroSent_to_oneSent ();
  spectre_d1_zeroSent_to_any ();
  spectre_d1_any_to_oneSent ();
  spectre_d1_any_to_any ();
  spectre_d1_up ();
  spectre_update_zeroSent_to_oneSent_2 ();
  spectre_d2_up ();
  spectre_d2_zeroSent_to_oneSent ();
  spectre_d2_zeroSent_to_any ();
  spectre_d2_any_to_oneSent ();
  spectre_d2_any_to_any ();
  spectre_current_location = spectre_location_oneSent;
}

void
spectre_transition_zeroSent_to_idle_3 (void)
{
  spectre_d1_zeroSent_to_idle ();
  spectre_d1_zeroSent_to_any ();
  spectre_d1_any_to_idle ();
  spectre_d1_any_to_any ();
  spectre_update_zeroSent_to_idle_3 ();
  spectre_d2_zeroSent_to_idle ();
  spectre_d2_zeroSent_to_any ();
  spectre_d2_any_to_idle ();
  spectre_d2_any_to_any ();
  spectre_current_location = spectre_location_idle;
}

void
spectre_transition_zeroSent_to_idle_4 (void)
{
  spectre_d1_zeroSent_to_idle ();
  spectre_d1_zeroSent_to_any ();
  spectre_d1_any_to_idle ();
  spectre_d1_any_to_any ();
  spectre_update_zeroSent_to_idle_4 ();
  spectre_d2_zeroSent_to_idle ();
  spectre_d2_zeroSent_to_any ();
  spectre_d2_any_to_idle ();
  spectre_d2_any_to_any ();
  spectre_current_location = spectre_location_idle;
}

void
spectre_transition_waitingZero_to_zeroSent_1 (void)
{
  spectre_d1_waitingZero_to_zeroSent ();
  spectre_d1_waitingZero_to_any ();
  spectre_d1_any_to_zeroSent ();
  spectre_d1_any_to_any ();
  spectre_d1_down ();
  spectre_update_waitingZero_to_zeroSent_1 ();
  spectre_d2_down ();
  spectre_d2_waitingZero_to_zeroSent ();
  spectre_d2_waitingZero_to_any ();
  spectre_d2_any_to_zeroSent ();
  spectre_d2_any_to_any ();
  spectre_current_location = spectre_location_zeroSent;
}

/* location implementation -------------------------------------------------- */
void
spectre_location_idle (void)
{
  if (spectre_check_idle_to_oneSent_1 ())
    spectre_transition_idle_to_oneSent_1 ();
}

void
spectre_location_oneSent (void)
{
  if (spectre_check_oneSent_to_waitingOne_1 ())
    spectre_transition_oneSent_to_waitingOne_1 ();
  if (spectre_check_oneSent_to_zeroSent_2 ())
    spectre_transition_oneSent_to_zeroSent_2 ();
  if (spectre_check_oneSent_to_idle_3 ())
    spectre_transition_oneSent_to_idle_3 ();
}

void
spectre_location_waitingOne (void)
{
  if (spectre_check_waitingOne_to_oneSent_1 ())
    spectre_transition_waitingOne_to_oneSent_1 ();
}

void
spectre_location_zeroSent (void)
{
  if (spectre_check_zeroSent_to_waitingZero_1 ())
    spectre_transition_zeroSent_to_waitingZero_1 ();
  if (spectre_check_zeroSent_to_oneSent_2 ())
    spectre_transition_zeroSent_to_oneSent_2 ();
  if (spectre_check_zeroSent_to_idle_3 ())
    spectre_transition_zeroSent_to_idle_3 ();
  if (spectre_check_zeroSent_to_idle_4 ())
    spectre_transition_zeroSent_to_idle_4 ();
}

void
spectre_location_waitingZero (void)
{
  if (spectre_check_waitingZero_to_zeroSent_1 ())
    spectre_transition_waitingZero_to_zeroSent_1 ();
}

void
spectre_check_for_events (void)
{
}


void
spectre_cleanup (void)
{
#ifdef DEBUG
  rt_printk (" Stopping Sender Controller\n");
#endif
}

void
spectre_startup (void)
{
#ifdef DEBUG
  rt_printk (" Starting Sender Controller\n");
#endif
  var_p = 0;
  outb (0, CTRL_REGISTER);	//set port to output mode
}

int
spectre_equals (RTIME a, RTIME b)
{
  RTIME lower = b - WIDENER;
  RTIME higher = b + WIDENER;
  return (a > lower && a < higher);
}

int
spectre_largerthan (RTIME a, RTIME other)
{
  other -= WIDENER;
  return (a > other);
}

int
spectre_smallerthan (RTIME a, RTIME other)
{
  other += WIDENER;
  return (a < other);
}

/* main function implementation --------------------------------------------- */
void
spectre_task_main (int data)
{
  spectre_running = 1;
  spectre_startup ();
  spectre_transition_initially ();
  while (spectre_running == 1)
    {
      spectre_current_time = rt_get_time ();
      spectre_check_for_events ();
      spectre_current_location ();
      rt_task_wait_period ();
    }
  spectre_cleanup ();
}

/* kernel module functions -------------------------------------------------- */
int
spectre_init_module (void)
{
  rt_set_periodic_mode ();
  start_rt_timer (BEAT);
  rt_task_init (&spectre_task, spectre_task_main, 0, STACK_SIZE,
		TASK_PRIORITY, USE_FPU, 0);
  rt_task_make_periodic (&spectre_task, rt_get_time () + PERIOD, PERIOD);
  return 0;
}

void
spectre_cleanup_module (void)
{
  stop_rt_timer ();
  rt_task_delete (&spectre_task);
}

module_init (spectre_init_module);
module_exit (spectre_cleanup_module);
MODULE_LICENSE ("GPL");
