{
module Main (main) where
import Tokens
}

%wrapper "posn"

$digit = 0-9			-- digits
$alpha = [a-zA-Z]		-- alphabetic characters	    

tokens :-
    $white+               ;
    "--".*				  ;
    \{\- ((~\-)|(\- ~\})|\n)* \-\}  ;
    any                   { \p s -> TokenAny (toLineCol p) }  
    nop                   { \p s -> TokenNop (toLineCol p) }  
    none                  { \p s -> TokenNone (toLineCol p) }  
    specification         { \p s -> TokenSpecification (toLineCol p) }  
    events                { \p s -> TokenEvents (toLineCol p) }
    internals             { \p s -> TokenInternals (toLineCol p) }
    orders                { \p s -> TokenOrders (toLineCol p) }
    decoration            { \p s -> TokenDecoration (toLineCol p) }
    initially             { \p s -> TokenInitially (toLineCol p) }
    location              { \p s -> TokenLocation (toLineCol p) }
    clocks                { \p s -> TokenClocks (toLineCol p) }
    vars                  { \p s -> TokenVars (toLineCol p) }
    global                { \p s -> TokenGlobal (toLineCol p) }
    startup               { \p s -> TokenStartup (toLineCol p) }
    cleanup               { \p s -> TokenCleanup (toLineCol p) }
    order                 { \p s -> TokenOrder (toLineCol p) }
    event                 { \p s -> TokenEvent (toLineCol p) }
    internal              { \p s -> TokenInternal (toLineCol p) }
    reading               { \p s -> TokenReading (toLineCol p) }
    writing               { \p s -> TokenWriting (toLineCol p) }
    restrict              { \p s -> TokenRestrict (toLineCol p) }
    to                    { \p s -> TokenTo (toLineCol p) }
    end                   { \p s -> TokenEnd (toLineCol p) }
    \(                    { \p s -> TokenLeftParenthesis (toLineCol p) }
    \)                    { \p s -> TokenRightParenthesis (toLineCol p) }
    :=                    { \p s -> TokenAssign (toLineCol p) }
    :                     { \p s -> TokenColon (toLineCol p) }
    \;                    { \p s -> TokenSemiColon (toLineCol p) }
    \,                    { \p s -> TokenComma (toLineCol p) }
    \<\=                  { \p s -> TokenSmallerThan (toLineCol p) }
    \>\=                  { \p s -> TokenLargerThan (toLineCol p) }
    \=                    { \p s -> TokenEquals (toLineCol p) }
    \-                    { \p s -> TokenMinus (toLineCol p) }
    \+                    { \p s -> TokenPlus (toLineCol p) }
    \*                    { \p s -> TokenTimes (toLineCol p) }
    \{                    { \p s -> TokenLeftCurlyBrace (toLineCol p) }
    \}                    { \p s -> TokenRightCurlyBrace (toLineCol p) }
    \/                    { \p s -> TokenDivide (toLineCol p) }
    \{\% ((~\%)|(\% ~\})|\n)* \%\} { \p s -> TokenCCode (toLineCol p) (treatCode s) }
    $alpha [$alpha $digit \_ \']* { \p s -> TokenIdentifier (toLineCol p) s }
    $digit+               { \p s -> TokenNumber (toLineCol p) (read s)}
                    
{
-- Each action has type :: AlexPosn -> String -> Token
              
toLineCol :: AlexPosn -> TokenPos
toLineCol (AlexPn a l c) = LineCol l c              
              
main = do
  s <- getContents
  print (nicScanTokens s)
  
nicScanTokens str = go (alexStartPos,'\n',str)
  where go inp@(pos,_,str) =
          case alexScan inp 0 of
                AlexEOF -> []
                AlexError inp'@(AlexPn _ l c , char , _) -> lexError l c char
                AlexSkip  inp' len     -> go inp'
                AlexToken inp' len act -> act pos (take len str) : go inp'
                
lexError line col char = error msg
  where msg =   "Lexical error at line " ++
                (show line) ++
                ", column " ++
                (show col) ++
                " (unexpected character '" ++ [char] ++ "')"

treatCode :: String -> String
treatCode s = reverse (drop 2 (reverse (drop 2 s)))
                
}
