module Tokens where

data TokenPos = LineCol !Int  -- line number
                        !Int  -- column number       
                deriving(Eq,Show,Read)
                    
data Token =
--  TokenSilent           TokenPos |
--  TokenLeaving          TokenPos |
--  TokenEntering         TokenPos |
--  TokenFrom             TokenPos |
------------------------------------
    TokenAny              TokenPos |
    TokenNop              TokenPos |
    TokenNone             TokenPos |
------------------------------------
    TokenSpecification    TokenPos |
    TokenGlobal           TokenPos |
    TokenEvents           TokenPos |
    TokenOrders           TokenPos |
    TokenInternals        TokenPos |
	TokenDecoration       TokenPos |
	TokenInitially        TokenPos |
	TokenLocation         TokenPos |
	TokenClocks    	      TokenPos |
	TokenVars             TokenPos |
	TokenStartup          TokenPos |
	TokenCleanup          TokenPos |     
    TokenOrder            TokenPos |
    TokenEvent            TokenPos |
    TokenInternal         TokenPos |
    TokenReading          TokenPos |
    TokenWriting          TokenPos |
    TokenRestrict         TokenPos |
    TokenTo               TokenPos |
    TokenEnd              TokenPos |
    TokenColon            TokenPos |
    TokenSemiColon        TokenPos |
    TokenComma            TokenPos |
    TokenAssign           TokenPos |
    TokenSmallerThan      TokenPos |
    TokenLargerThan       TokenPos |
    TokenEquals           TokenPos |
    TokenTimes            TokenPos | 
    TokenDivide           TokenPos |   
    TokenPlus             TokenPos |
    TokenMinus            TokenPos |
    TokenNumber TokenPos Int       |
    TokenCCode  TokenPos String    |
    TokenLeftParenthesis  TokenPos |
    TokenRightParenthesis TokenPos |
    TokenLeftCurlyBrace   TokenPos |
    TokenRightCurlyBrace  TokenPos |  
    TokenIdentifier TokenPos String|
    BadToken
	deriving (Eq,Show,Read)
	
idName (TokenIdentifier pos name) = name
idLine i@(TokenIdentifier _ _) = let (LineCol l _) = tokenPos i in l
idCol i@(TokenIdentifier _ _) = let (LineCol _ c) = tokenPos i in c

codeContents (TokenCCode pos code) = code
numberValue (TokenNumber pos i) = i
	
tokenPos :: Token -> TokenPos
-- tokenPos (TokenEntering t) = t         
-- tokenPos (TokenFrom t) = t             
-- tokenPos (TokenSilent t) = t           
-- tokenPos (TokenLeaving t) = t          
----------------------------------
tokenPos (TokenAny t) = t
tokenPos (TokenNop t) = t
tokenPos (TokenNone t) = t
----------------------------------
tokenPos (TokenSpecification t) = t 
tokenPos (TokenEvents t) = t 
tokenPos (TokenOrders t) = t 
tokenPos (TokenInternals t) = t        
tokenPos (TokenGlobal t) = t     
tokenPos (TokenDecoration t) = t       
tokenPos (TokenInitially t) = t        
tokenPos (TokenLocation t) = t         
tokenPos (TokenClocks t) = t     	    
tokenPos (TokenVars t) = t             
tokenPos (TokenStartup t) = t          
tokenPos (TokenCleanup t) = t               
tokenPos (TokenOrder t) = t            
tokenPos (TokenEvent t) = t            
tokenPos (TokenInternal t) = t         
tokenPos (TokenReading t) = t          
tokenPos (TokenWriting t) = t          
tokenPos (TokenRestrict t) = t         
tokenPos (TokenTo t) = t               
tokenPos (TokenEnd t) = t              
tokenPos (TokenColon t) = t            
tokenPos (TokenSemiColon t) = t        
tokenPos (TokenComma t) = t           
tokenPos (TokenAssign t) = t          
tokenPos (TokenSmallerThan t) = t      
tokenPos (TokenLargerThan t) = t       
tokenPos (TokenEquals t) = t           
tokenPos (TokenTimes t) = t            
tokenPos (TokenDivide t) = t             
tokenPos (TokenPlus t) = t             
tokenPos (TokenMinus t) = t            
tokenPos (TokenNumber t _ ) = t
tokenPos (TokenCCode t _ ) = t
tokenPos (TokenLeftParenthesis t) = t
tokenPos (TokenRightParenthesis t) = t
tokenPos (TokenLeftCurlyBrace t) = t
tokenPos (TokenRightCurlyBrace t) = t
tokenPos (TokenIdentifier t _ ) = t

tokenStr :: Token -> String
-- tokenStr (TokenEntering t) = "entering"  
-- tokenStr (TokenFrom t) = "from"
-- tokenStr (TokenSilent t) = "silent"
-- tokenStr (TokenLeaving t) = "leaving"   
-------------------------------------------
tokenStr (TokenAny t) = "any"
tokenStr (TokenNop t) = "nop"
tokenStr (TokenNone t) = "none"
-------------------------------------------
tokenStr (TokenSpecification t) = "specification"
tokenStr (TokenEvents t) = "events" 
tokenStr (TokenOrders t) = "orders"
tokenStr (TokenGlobal t) = "global"
tokenStr (TokenInternals t) = "internals"    
tokenStr (TokenDecoration t) = "decoration"    
tokenStr (TokenInitially t) = "initially"
tokenStr (TokenLocation t) = "location"
tokenStr (TokenClocks t) = "clocks"  
tokenStr (TokenVars t) = "vars"    
tokenStr (TokenStartup t) = "startup"
tokenStr (TokenCleanup t) = "cleanup"       
tokenStr (TokenOrder t) = "order"   
tokenStr (TokenEvent t) = "event"      
tokenStr (TokenInternal t) = "internal"      
tokenStr (TokenReading t) = "reading"
tokenStr (TokenWriting t) = "writing"          
tokenStr (TokenRestrict t) = "restrict"
tokenStr (TokenTo t) = "to"        
tokenStr (TokenEnd t) = "end"             
tokenStr (TokenColon t) = ":"          
tokenStr (TokenSemiColon t) = ";"
tokenStr (TokenComma t) = ","     
tokenStr (TokenAssign t) = ":="         
tokenStr (TokenSmallerThan t) = "<="
tokenStr (TokenLargerThan t) = ">="   
tokenStr (TokenEquals t) = "="    
tokenStr (TokenTimes t) = "*"          
tokenStr (TokenDivide t) = "/"            
tokenStr (TokenPlus t) = "+"             
tokenStr (TokenMinus t) = "-"           
tokenStr (TokenNumber t _ ) = "number"
tokenStr (TokenCCode t _ ) = "code fragment"
tokenStr (TokenLeftParenthesis t) = "("
tokenStr (TokenRightParenthesis t) = ")"
tokenStr (TokenLeftCurlyBrace t) = "{"
tokenStr (TokenRightCurlyBrace t) = "}"
tokenStr (TokenIdentifier t _ ) = "identifier"
